import asyncio
import functools
import json
import sys
from urllib.parse import quote

import aiohttp

from .errors import HttpException, Forbidden, LoginFailure, NotFound, EmoteExists
from .utils import sentinel
from . import __version__

# Using code provided by Rapptz
# Copyright © 2015–2017 Rapptz
# https://github.com/Rapptz/discord.py/blob/4aecdea0524e7b481f9750166bf9e9be287ec445/discord/http.py

# by default, quote doesn't quote /, which we don't want.
# quote_plus does, but it also encodes " " as "+", which we don't want.
uriquote = functools.partial(quote, safe='')
del quote

async def json_or_text(response):
	text = await response.text(encoding='utf-8')
	if response.content_type == 'application/json':
		return json.loads(text)
	return text

class Route:
	BASE = 'https://emote-collector.python-for.life/api/v0'

	def __init__(self, method, path, **parameters):
		self.path = path
		self.method = method
		url = (self.BASE + self.path)
		if parameters:
			self.url = url.format(**{k: uriquote(v) if isinstance(v, str) else v for k, v in parameters.items()})
		else:
			self.url = url

class HttpClient:
	def __init__(self, token=None, *, loop=None):
		self.token = token
		self.loop = loop or asyncio.get_event_loop()
		user_agent = 'aioec (https://github.com/bmintz/aioec) {0} aiohttp/{1} Python/{2[0]}.{2[1]}'
		self.user_agent = user_agent.format(__version__, aiohttp.__version__, sys.version_info)

		headers = {'User-Agent': self.user_agent}
		if self.token is not None:
			headers['Authorization'] = self.token

		self._session = aiohttp.ClientSession(headers=headers, loop=self.loop)

	def close(self):
		return self._session.close()

	async def request(self, route, **kwargs):
		method = route.method
		url = route.url

		async with self._session.request(method, url, **kwargs) as response:
			data = await json_or_text(response)
			if response.status in range(200, 300):
				return data

			if response.status == 401:
				raise LoginFailure
			elif response.status == 403:
				raise Forbidden(response, data)
			elif response.status == 404:
				raise NotFound(response, data)
			elif response.status == 409:  # Conflict
				raise EmoteExists(response, data['name'])
			else:
				raise HttpException(response, data)

	def emotes(self):
		return self.request(Route('GET', '/emotes'))

	def emote(self, name):
		return self.request(Route('GET', '/emote/{name}', name=name))

	def login(self):
		return self.request(Route('GET', '/login'))

	def create(self, name, url):
		return self.request(Route('PUT', '/emote/{name}/{url}', name=name, url=url))

	def edit(self, name_, *, name=None, description=sentinel):
		data = {}

		# we perform this dance so that the caller can run it like edit_emote('foo', name='bar')
		new_name = name
		name = name_

		if new_name is not None:
			data['name'] = new_name
		if description is not sentinel:  # None is an allowed value for description
			data['description'] = description

		return self.request(Route('PATCH', '/emote/{name}', name=name), json=data)

	def delete(self, name):
		return self.request(Route('DELETE', '/emote/{name}', name=name))

	def search(self, query):
		return self.request(Route('GET', '/search/{query}', query=query))

	def popular(self):
		return self.request(Route('GET', '/popular'))



from . import utils

class BaseEmote:
	_private_slots_ = frozenset(('_data',))

	__slots__ = _private_slots_ | frozenset((
		'_name',
		'_id',
		'_author',
		'_animated',
		'_created',
		'_modified',
		'_preserve',
		'_description',
		'_usage',
	))

	def __new__(cls, *, data):
		self = super().__new__(cls)
		self._data = data

		for key, value in data.items():
			if key in {'id', 'author'}:
				value = int(value)

			if key in {'created', 'modified'} and value is not None:
				value = utils.epoch_time(value)

			setattr(self, '_' + key, value)

		return self

	for field in __slots__:
		if field in _private_slots_:
			continue

		def getter(self, field=field):
			return getattr(self, field)

		public_name = field[1:]
		getter.__name__ = public_name
		vars()[public_name] = property(getter)
		del getter, public_name

	del field

	@property
	def usage(self):
		return getattr(self, '_usage', None)

	@property
	def url(self):
		return 'https://cdn.discordapp.com/emojis/{0.id}.{ext}?v=1'.format(
			self,
			ext='gif' if self.animated else 'png')

	@property
	def _a(self):
		return 'a' if self.animated else ''

	@property
	def as_reaction(self):
		return '{0._a}:{0.name}:{0.id}'.format(self)

	def __str__(self):
		return '<{0._a}:{0.name}:{0.id}>'.format(self)

	def __repr__(self):
		return (
			'{0.__module__}.{0.__class__.__qualname__}'
			'<name={0.name!r}, id={0.id}, animated={0.animated}>'.format(self))

class Emote(BaseEmote):
	_private_slots_ = frozenset(('_http',))
	__slots__ = _private_slots_

	def __new__(cls, *, data, http):
		self = super().__new__(cls, data=data)
		self._http = http

		return self

	def _new_emote(self, data):
		return type(self)(http=self._http, data=data)

	async def delete(self):
		return self._new_emote(await self._http.delete(self.name))

	async def edit(self, *, name=None, description=utils.sentinel):
		return self._new_emote(await self._http.edit(self.name, name=name, description=description))



# Using code provided by Rapptz
# Copyright © 2015–2017 Rapptz
# https://github.com/Rapptz/discord.py/blob/f25091efe1281aebe70189c61f9cac405b21a72f/discord/errors.py

class AioEcError(Exception):
	"""
	Base exception type for the library.
	Can be used to catch any exception raised by this library.
	"""
	pass

class ClientException(AioEcError):
	"""Exception that's thrown when an operation in the :class:`Client` fails.

	These are usually for exceptions that happened due to user input.
	"""
	pass

class LoginFailure(ClientException):
	"""Improper or incorrect token has been passed"""
	def __init__(self):
		super().__init__('Invalid or incorrect token has been passed.')

class HttpException(AioEcError):
	"""Exception that's thrown when an HTTP request operation fails.

	Attributes
	------------
	response: aiohttp.ClientResponse
		The response of the failed HTTP request. This is an
		instance of `aiohttp.ClientResponse`__. In some cases
		this could also be a ``requests.Response``.

		__ http://aiohttp.readthedocs.org/en/stable/client_reference.html#aiohttp.ClientResponse

	text: :class:`str`
		The text of the error. Could be an empty string.
	status: :class:`int`
		The status code of the HTTP request.
	code: :class:`int`
		The Discord specific error code for the failure.
	"""

	def __init__(self, response, message):
		self.response = response
		self.status = response.status
		if isinstance(message, dict):
			self.text = message.get('message')
		else:
			self.text = message

		fmt = '{0.reason} (status code: {0.status})'
		if self.text:
			fmt = fmt + ': {1}'

		super().__init__(fmt.format(self.response, self.text))

class Forbidden(HttpException):
	"""Exception that's thrown for when status code 403 occurs.

	Subclass of :exc:`HttpException`
	"""
	pass

class NotFound(HttpException):
	"""Exception that's thrown for when status code 404 occurs.

	Subclass of :exc:`HttpException`
	"""
	pass

class EmoteExists(HttpException):
	"""Exception that's thrown for when status code 404 occurs.
	This happens when an emote already exists with the given name and you tried to create a new one
	with that name.

	Subclass of :exc:`HttpException`
	"""
	def __init__(self, response, name):
		self.name = name
		super().__init__(response, 'An emote called {} already exists'.format(name))



#!/usr/bin/env python3
# encoding: utf-8

import re
import setuptools

with open('aioec/__init__.py') as f:
	version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)

with open('README.rst') as f:
	readme = f.read()

setuptools.setup(
	name='aioec',
	author='Benjamin Mintz',
	author_email='bmintz@protonmail.com',
	url='https://github.com/EmoteCollector/aioec',
	version=version,
	license='MIT',
	packages=['aioec'],
	install_requires=['aiohttp>=3.3.0,<3.4.0'],
	description='async client library for the Emote Collector API',
	long_description=readme,
	classifiers=[
		'Development Status :: 2 - Pre-Alpha',
		'Intended Audience :: Developers',
		'Topic :: Internet',
		'Topic :: Software Development :: Libraries',
		'Topic :: Software Development :: Libraries :: Python Modules',
		'Programming Language :: Python :: 3.5',
		'Programming Language :: Python :: 3.6',
		'Programming Language :: Python :: 3.7',
		'Operating System :: OS Independent',
		'License :: OSI Approved :: MIT License',
	],
)



from .http import HttpClient
from .emote import BaseEmote, Emote
from . import utils

class Client:
	def __init__(self, token=None, *, loop=None):
		self._http = HttpClient(token=token, loop=loop)

	def _new_emote(self, data):
		return Emote(data=data, http=self._http)

	def close(self):
		return self._http.close()

	async def emote(self, name):
		return self._new_emote(await self._http.emote(name))

	async def emotes(self):
		return map(self._new_emote, await self._http.emotes())

	async def search(self, query):
		return map(self._new_emote, await self._http.search(query))

	async def popular(self):
		return map(self._new_emote, await self._http.popular())

	async def login(self):
		"""Checks that your token is correct.

		Returns the user ID associated with your token.
		"""

		return int(await self._http.login())

	async def create(self, name, url):
		return self._new_emote(await self._http.create(name, url))

	async def edit(self, name_, *, name=None, description=utils.sentinel):
		return self._new_emote(await self._http.edit(name_, name=name, description=description))

	async def delete(self, name):
		return self._new_emote(await self._http.delete(name))

	async def __aenter__(self):
		return self

	def __aexit__(self, *_):
		return self.close()
